# FFTSearch: Genarating and Searching Families of FFT Algorithms
# Copyright (C) 2011 Steve Haynal, Heidi Haynal
#
# This file is part of FFTSearch.
#
# FFTSearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFTSearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

## Support for modular arithmetic
class ModInt:
	def __init__(self,v,mod):
		self.mod = mod
		self.v = v % self.mod
		
	def __repr__(self):
		return "ModInt(%d,%d)" % (self.v,self.mod)
		
	def __str__(self):
		return str(self.v)
		
	def NMod(self,mod):
		if mod > self.mod:
			m = mod / self.mod
			self.mod = mod
			self.v = self.v * m
		elif mod < self.mod:
			d = self.mod /mod
			self.mod = mod
			self.v = self.v / d
		return self
		
	def __add__(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return ModInt((self.v+other.v),self.mod)
		
	def __sub__(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return ModInt((self.v-other.v),self.mod)

	def __lshift__(self,other):
		return ModInt((self.v << other),self.mod)
		
	def __rshift__(self,other):
		return ModInt((self.v >> other),self.mod)

	def __eq__(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return self.v == other.v
		
	def __ne__(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return self.v != other.v		
		
	def __lt__(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return self.v < other.v
		
	def __le__(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return self.v <= other.v	
			
	def __gt__(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return self.v > other.v
		
	def __ge__(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return self.v >= other.v	
		
	def __int__(self):
		return self.v
		
	def __neg__(self):
		return ModInt(self.mod - self.v,self.mod)
		
	def __invert__(self):
		return ModInt(self.v + (self.mod/2),self.mod)
		
	def __mod__(self,newmod):
		return ModInt(self.v,newmod)
		
	def Norm(self):
		return ModInt(self.v & ((self.mod-1) >> 2),self.mod)
		
	def EqualRealImag(self):
		return ((self.v & ((self.mod-1) >> 3)) == 0) and (self.v & ((self.mod-1) >> 2))
		
	def Conjugate(self):
		return ModInt(self.mod - self.v,self.mod)
		
	def Cost(self):
		mask = self.mod -1
		ntf = self.v & (mask >> 2)
		if ntf:
			ntf = self.v & (mask >> 3)
			if ntf:
				return 6
			else:
				return 4
		else:
			return 0
		
	def IsConjugate(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return ((self.v + other.v) % self.mod) == 0
		
	def IsNormConjugate(self,other):
		assert self.mod == other.mod, 'ERROR: Arithmetic with differing mod.'
		return ((self.v + other.v) & ((self.mod-1) >> 2)) == 0		
		
		
		
		
		
