# FFTSearch: Genarating and Searching Families of FFT Algorithms
# Copyright (C) 2011 Steve Haynal, Heidi Haynal
#
# This file is part of FFTSearch.
#
# FFTSearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFTSearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from BitVector import *

## Node in FFT flowgraph
class Node:
	def __init__(self,stride,base,Wstride,Wbase=None):
		
		self.stride		= stride
		self.base		= base				
		self.Wstride	= Wstride

		self.id		= (self.stride,self.base,self.Wstride.v) 
		self.hid	= hash(self.id)
		
		self.tpfg = None
		self.bpfg = None

		self.lc = None
		self.rc = None
		self.lp = None
		self.rp = None
		
		self.ltfp = None
		self.rtfp = None

		self.Wbase = Wbase
		
		self.i = None

	def __cmp__(self,other):
		if not isinstance(other,Node): 
			return -1
		else:
			res = cmp(self.stride,other.stride)
			if res != 0: return res
			
			res = cmp(self.base,other.base)
			if res != 0: return res
			
			## Reverse bit sort on Wstride
			lb = self.Wstride.v
			rb = other.Wstride.v
			while lb != rb:
				if (lb & 1) < (rb & 1): return -1
				elif (lb & 1) > (rb & 1): return 1
				lb = lb >> 1
				rb = rb >> 1 

			return 0
		
	def __hash__(self):
		return self.hid		

	def __str__(self):
		return '"%d:%d:%d"' % self.id
		
	def __repr__(self):
		return "Node(%d,%d,%d)" % self.id
		
	## To match Flowgraph
	def GetID(self): return self.id
	def SubFlowGraphs(self): return iter([])
	def InpButterflies(self): return iter([])
	def Inps(self): return iter([self])

	def GetTFOffset(self): return self.rtfp - self.ltfp
	
	def SetTF(self,tf):
		tfoffset = self.GetTFOffset()
		self.ltfp = tf
		self.rtfp = tf + tfoffset
		
	def rWbase(self): return self.Wbase + self.Wstride
		
	def IsOriginalAncestor(self): return self.lp == None
	def IsTerminalDescendent(self): return self.lc == None
		
	def Cost(self):

		flops = self.ltfp.Cost()
		
		## Add extra cost for rtfp if not same normalized value
		if self.ltfp.Norm() != self.rtfp.Norm():
			## Only 2 additional flops if conjugate
			if self.ltfp.IsNormConjugate(self.rtfp):
				flops += 2
			else:
				flops += self.rtfp.Cost()
		return flops
		
	def Cost6(self):
		c = self.Cost()
		if c == 6:
			return 6
		else:
			return 0
			
		
	def Name(self,prefix=""):
		sid = "%d_%d_%d" % self.id
		return "%s_%s" % (prefix,sid)

	
