# FFTSearch: Genarating and Searching Families of FFT Algorithms
# Copyright (C) 2011 Steve Haynal, Heidi Haynal
#
# This file is part of FFTSearch.
#
# FFTSearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFTSearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from Flowgraph import *
import commands

## Coordinating class for FFT searches
class SearchTemplate:
	
	def __init__(self,m,offset):
		
		self.m = m
		self.offset = offset
		
		print "Build flowgraph...",
		self.fg = Flowgraph(self.m)
		self.fg.InitTopFlowgraph()
		print "Done"
		
		print "Set initial flowgraph algorithm...",
		self.ffta = FFTAlgorithm(self.m)
		self.ffta.SplitRadix()
		self.fg.SetFFTAlgorithm(self.ffta)
		## Apply transformations so no cost in first or last two rows
		self.fg.PrepareDIT()
		self.fg.UpdateWeightsFromTFs()
		self.ffta = self.fg.ExtractFFTAlgorithm()
		print "Done"
		
		print "Validate and compute cost of initial flowgraph..."
		self.fg.Validate()
		print "Flowgraph cost is",self.fg.Cost(),"flops."
		print "Done\n"

	def PartitionSMTSearch(self,smt):
		smt.DoMaxCost()
		
		
	def DoSMTSearch(self):
		'''Search for FFTs of specified cost'''
		print "Starting SMT Search using",len(self.partitions),"partitions"
		tcost = 0	 
		i = 0	
		sat = 0
		for (d,sfg) in self.partitions:
			cost = sfg.Cost(False)
			costlimit = cost - (d + self.offset)
			tcost = tcost + costlimit
			if costlimit < 0:
				costlimit = 0
			rootname = "%s_%d" % (self.name,i)
			print "Subflowgraph size",sfg.N,"has original cost of",cost,"with new cost limit",costlimit,
			smt = BFConstraints(sfg,costlimit+(2*sfg.N*sfg.m))
			self.PartitionSMTSearch(smt)
			print "Rootname is",rootname
			smt.WriteSMTLib1_2("%s.smt" % rootname)
			res = commands.getoutput('time boolector -fm -o %s.txt %s.smt' % (rootname,rootname))
			print res
			f = open("%s.txt" % rootname,'r')
			l = f.readline()
			if l[:3] == "sat":
				sat = sat + 1
			print l
			f.close()
			i += 1		
		
		print "Total attempted flop count is",tcost+(2*self.fg.N*self.fg.m)		
		if sat == len(self.partitions):
			print "All partitions are SAT"
			self.CreateAlgorithmInstance()
		elif sat == 0:
			print "All partitions are UNSAT"
		else:
			print sat,"out of",len(self.partitions),"are SAT"


	def DoLIASearch(self):
		'''Search for FFTs of specified cost'''
		print "Starting SMT Search using",len(self.partitions),"partitions"
		tcost = 0	 
		i = 0	
		sat = 0
		for (d,sfg) in self.partitions:
			cost = sfg.Cost(False)
			costlimit = cost - (d + self.offset)
			tcost = tcost + costlimit
			if costlimit < 0:
				costlimit = 0
			rootname = "%s_%d" % (self.name,i)
			print "Subflowgraph size",sfg.N,"has original cost of",cost,"with new cost limit",costlimit,
			smt = BFConstraints(sfg,costlimit+(2*sfg.N*sfg.m))
			self.PartitionSMTSearch(smt)
			print "Rootname is",rootname
			smt.WriteLIA("%s.smt2" % rootname)
			i += 1		
		
		print "Total attempted flop count is",tcost+(2*self.fg.N*self.fg.m)		



	def DoQFBVSearch(self):
		'''Search for FFTs of specified cost'''
		print "Starting SMT Search using",len(self.partitions),"partitions"
		tcost = 0	 
		i = 0	
		sat = 0
		for (d,sfg) in self.partitions:
			cost = sfg.Cost(False)
			costlimit = cost - (d + self.offset)
			tcost = tcost + costlimit
			if costlimit < 0:
				costlimit = 0
			rootname = "%s_%d" % (self.name,i)
			print "Subflowgraph size",sfg.N,"has original cost of",cost,"with new cost limit",costlimit,
			smt = BFConstraints(sfg,costlimit+(2*sfg.N*sfg.m))
			self.PartitionSMTSearch(smt)
			print "Rootname is",rootname
			smt.WriteQFBV("%s.smt2" % rootname)
			i += 1		
		
		print "Total attempted flop count is",tcost+(2*self.fg.N*self.fg.m)		



	def CreateAlgorithmInstance(self):
		'''Create concrete FFT instance'''
		print "Creating final SMT result file %s.txt" % self.name
		## Create final result file
		fo = open("%s.txt" % self.name,'w')
			
		for i in range(0,len(self.partitions)):
			f = open("%s_%d.txt" % (self.name,i))
			fo.write(f.read())
		fo.close()
			
		print "Creating FFT algorithm instance test file %s.py" % self.name
		self.ffta.ParseBoolectorBFResult("%s.txt" % self.name)			

		self.fg.SetFFTAlgorithm(self.ffta)
		print "Final instance cost is",self.fg.Cost(),"flops."
		self.fg.Validate()		
		
		ofn = open("%s.py" % self.name,'w')
		ofn.write("from TestFFT import *\n\n")
	
		ofn.write("t=TestFFT(%d)\n" % self.fg.N)
	
		lstride = 0
		lwstride = 0
		row = 0
		for bf in self.fg.Butterflies():
			if bf.tl.stride != lstride:
				lstride = bf.tl.stride
				lwstride = bf.tl.Wstride.v
				ofn.write("\n\n#### New Stride %d\n" % lstride)
				ofn.write("######## New Weight Stride %d\n" % lwstride)			
			elif bf.tl.Wstride.v != lwstride:
				lwstride = bf.tl.Wstride.v 
				ofn.write("\n######## New Weight Stride %d\n" % lwstride)			

			ofn.write("t.BF(%d,%d,%d,%d)\n" % (bf.tl.i,bf.tr.i,bf.tl.ltfp.v,bf.tr.ltfp.v))
		
		done = False
		for nd in self.fg.outs:
			if nd.ltfp.v != 0:
				if not done:
					ofn.write("#### Final weight applied to some outputs\n")
					done = True
				ofn.write("t.FW(%d,%d)\n" % (nd.i,nd.ltfp.v))
		
		ofn.write("\n\nt.Report()\n")
		ofn.close()
		
		print "Executing instance script"
		execfile("%s.py" % self.name)

	
	
class Sz32_456(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,5,offset)
		self.name = "Sz32_456"	

		self.partitions = [
			(0,self.fg.br.br), 
			(0,self.fg.br.bl),
#	 
			(0,self.fg.bl.br),
			(0,self.fg.bl.bl)]	


class Sz64_1160(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,6,offset)
		self.name = "Sz64_1160"	

		self.partitions = [
			(0,self.fg.br.br.tr.tr), 
			(0,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(0,self.fg.br.bl.tr.tr),
			(0,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr),
			(0,self.fg.bl.br.br.tl),
			(0,self.fg.bl.br.bl.tr),
			(0,self.fg.bl.br.bl.tl),
#	 
			(0,self.fg.bl.bl.br),
			(0,self.fg.bl.bl.bl)]


class Sz128_2824(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,7,offset)
		self.name = "Sz128_2824"	

		self.partitions = [
			(0,self.fg.br.br.tr.tr), 
			(0,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(0,self.fg.br.bl.tr.tr),
			(0,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr),
			(0,self.fg.bl.br.br.tl),
			(0,self.fg.bl.br.bl.tr),
			(0,self.fg.bl.br.bl.tl),
#	 
			(0,self.fg.bl.bl.br),
			(0,self.fg.bl.bl.bl)]



class Sz256_6616(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,8,offset)
		self.name = "Sz256_6616"
		
		self.partitions = [
			(12,self.fg.br.br.tr.tr), 
			(12,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(12,self.fg.br.bl.tr.tr),
			(12,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr.tr),
			(0,self.fg.bl.br.br.tr.tl),	 
			(0,self.fg.bl.br.br.tl.tr),
			(0,self.fg.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.br.bl.tr.tr),
			(0,self.fg.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.br.bl.tl.tr),
			(0,self.fg.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.br.br.tr.tr),
			(0,self.fg.bl.bl.br.br.tr.tl),	 
			(0,self.fg.bl.bl.br.br.tl.tr),
			(0,self.fg.bl.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.bl.br.bl.tr.tr),
			(0,self.fg.bl.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.bl.br.bl.tl.tr),
			(0,self.fg.bl.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.bl.br.br.tr),
			(0,self.fg.bl.bl.bl.br.br.tl),
			(0,self.fg.bl.bl.bl.br.bl.tr),
			(0,self.fg.bl.bl.bl.br.bl.tl),	 
#	 
			(0,self.fg.bl.bl.bl.bl.br),
			(0,self.fg.bl.bl.bl.bl.bl)]
			
		
class Sz256_6664(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,8,offset)
		self.name = "Sz256_6664"
		
		self.partitions = [
			(0,self.fg.br.br.tr.tr), 
			(0,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(0,self.fg.br.bl.tr.tr),
			(0,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr.tr),
			(0,self.fg.bl.br.br.tr.tl),	 
			(0,self.fg.bl.br.br.tl.tr),
			(0,self.fg.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.br.bl.tr.tr),
			(0,self.fg.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.br.bl.tl.tr),
			(0,self.fg.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.br.br.tr.tr),
			(0,self.fg.bl.bl.br.br.tr.tl),	 
			(0,self.fg.bl.bl.br.br.tl.tr),
			(0,self.fg.bl.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.bl.br.bl.tr.tr),
			(0,self.fg.bl.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.bl.br.bl.tl.tr),
			(0,self.fg.bl.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.bl.br.br.tr),
			(0,self.fg.bl.bl.bl.br.br.tl),
			(0,self.fg.bl.bl.bl.br.bl.tr),
			(0,self.fg.bl.bl.bl.br.bl.tl),	 
#	 
			(0,self.fg.bl.bl.bl.bl.br),
			(0,self.fg.bl.bl.bl.bl.bl)]	
		
		
class Sz256_6632(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,8,offset)
		self.name = "Sz256_6632"
		
		self.partitions = [
			(8,self.fg.br.br.tr.tr), 
			(8,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(8,self.fg.br.bl.tr.tr),
			(8,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr.tr),
			(0,self.fg.bl.br.br.tr.tl),	 
			(0,self.fg.bl.br.br.tl.tr),
			(0,self.fg.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.br.bl.tr.tr),
			(0,self.fg.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.br.bl.tl.tr),
			(0,self.fg.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.br.br.tr.tr),
			(0,self.fg.bl.bl.br.br.tr.tl),	 
			(0,self.fg.bl.bl.br.br.tl.tr),
			(0,self.fg.bl.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.bl.br.bl.tr.tr),
			(0,self.fg.bl.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.bl.br.bl.tl.tr),
			(0,self.fg.bl.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.bl.br.br.tr),
			(0,self.fg.bl.bl.bl.br.br.tl),
			(0,self.fg.bl.bl.bl.br.bl.tr),
			(0,self.fg.bl.bl.bl.br.bl.tl),	 
#	 
			(0,self.fg.bl.bl.bl.bl.br),
			(0,self.fg.bl.bl.bl.bl.bl)]

		
class Sz512_15128(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,9,offset)
		self.name = "Sz512_15128"
		
		self.partitions = [
			(36,self.fg.br.br.tr.tr), 
			(36,self.fg.br.br.tr.tl),
			(24,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(36,self.fg.br.bl.tr.tr),
			(36,self.fg.br.bl.tr.tl),
			(24,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(12,self.fg.bl.br.br.tr.tr),
			(12,self.fg.bl.br.br.tr.tl),	 
			(0,self.fg.bl.br.br.tl.tr),
			(0,self.fg.bl.br.br.tl.tl),	 	 
			(12,self.fg.bl.br.bl.tr.tr),
			(12,self.fg.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.br.bl.tl.tr),
			(0,self.fg.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.br.br.tr.tr),
			(0,self.fg.bl.bl.br.br.tr.tl),	 
			(0,self.fg.bl.bl.br.br.tl.tr),
			(0,self.fg.bl.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.bl.br.bl.tr.tr),
			(0,self.fg.bl.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.bl.br.bl.tl.tr),
			(0,self.fg.bl.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.bl.br.br.tr.tr),
			(0,self.fg.bl.bl.bl.br.br.tr.tl),	 
			(0,self.fg.bl.bl.bl.br.br.tl.tr),
			(0,self.fg.bl.bl.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.bl.bl.br.bl.tr.tr),
			(0,self.fg.bl.bl.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.bl.bl.br.bl.tl.tr),
			(0,self.fg.bl.bl.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.bl.bl.br.br.tr),
			(0,self.fg.bl.bl.bl.bl.br.br.tl),
			(0,self.fg.bl.bl.bl.bl.br.bl.tr),
			(0,self.fg.bl.bl.bl.bl.br.bl.tl),	 
#	 
			(0,self.fg.bl.bl.bl.bl.bl.br),
			(0,self.fg.bl.bl.bl.bl.bl.bl)]				
		
		

class Sz512_15368(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,9,offset)
		self.name = "Sz512_15368"
		
		self.partitions = [
			(0,self.fg.br.br.tr.tr), 
			(0,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(0,self.fg.br.bl.tr.tr),
			(0,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr.tr),
			(0,self.fg.bl.br.br.tr.tl),	 
			(0,self.fg.bl.br.br.tl.tr),
			(0,self.fg.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.br.bl.tr.tr),
			(0,self.fg.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.br.bl.tl.tr),
			(0,self.fg.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.br.br.tr.tr),
			(0,self.fg.bl.bl.br.br.tr.tl),	 
			(0,self.fg.bl.bl.br.br.tl.tr),
			(0,self.fg.bl.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.bl.br.bl.tr.tr),
			(0,self.fg.bl.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.bl.br.bl.tl.tr),
			(0,self.fg.bl.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.bl.br.br.tr.tr),
			(0,self.fg.bl.bl.bl.br.br.tr.tl),	 
			(0,self.fg.bl.bl.bl.br.br.tl.tr),
			(0,self.fg.bl.bl.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.bl.bl.br.bl.tr.tr),
			(0,self.fg.bl.bl.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.bl.bl.br.bl.tl.tr),
			(0,self.fg.bl.bl.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.bl.bl.br.br.tr),
			(0,self.fg.bl.bl.bl.bl.br.br.tl),
			(0,self.fg.bl.bl.bl.bl.br.bl.tr),
			(0,self.fg.bl.bl.bl.bl.br.bl.tl),	 
#	 
			(0,self.fg.bl.bl.bl.bl.bl.br),
			(0,self.fg.bl.bl.bl.bl.bl.bl)]		
		
		



	
class Sz32_616rtfs(SearchTemplate):
	
	def __init__(self,offset):
		
		offset = offset - 160
	
		SearchTemplate.__init__(self,5,offset)
		self.name = "Sz32_616rtfs"	

		self.partitions = [
			(0,self.fg)]

	def PartitionSMTSearch(self,smt):
		constraints = [ConstVec(4,smt.width),ConstVec(5,smt.width)]
		smt.DoRestrictTFs(constraints)
		
	
	
class Sz64_2112rtfs(SearchTemplate):
	
	def __init__(self,offset):
		
		offset = offset - 952
	
		SearchTemplate.__init__(self,6,offset)
		self.name = "Sz64_2112rtfs"	

		self.partitions = [
			(0,self.fg)]

	def PartitionSMTSearch(self,smt):
		constraints = [ConstVec(7,smt.width),ConstVec(8,smt.width),ConstVec(9,smt.width)]
		smt.DoRestrictTFs(constraints)	





class Sz32_lia(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,5,offset)
		self.name = "Sz32_lia"	

		self.partitions = [
			(0,self.fg.br.br), 
			(0,self.fg.br.bl),
#	 
			(0,self.fg.bl.br),
			(0,self.fg.bl.bl)]	

	def PartitionSMTSearch(self,smt):
		smt.DoLIA()


class Sz64_lia(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,6,offset)
		self.name = "Sz64_lia"	

		self.partitions = [
			(0,self.fg.br.br.tr.tr), 
			(0,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(0,self.fg.br.bl.tr.tr),
			(0,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr),
			(0,self.fg.bl.br.br.tl),
			(0,self.fg.bl.br.bl.tr),
			(0,self.fg.bl.br.bl.tl),
#	 
			(0,self.fg.bl.bl.br),
			(0,self.fg.bl.bl.bl)]

	def PartitionSMTSearch(self,smt):
		smt.DoLIA()




class Sz128_lia(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,7,offset)
		self.name = "Sz128_lia"	

		self.partitions = [
			(0,self.fg.br.br.tr.tr), 
			(0,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(0,self.fg.br.bl.tr.tr),
			(0,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr),
			(0,self.fg.bl.br.br.tl),
			(0,self.fg.bl.br.bl.tr),
			(0,self.fg.bl.br.bl.tl),
#	 
			(0,self.fg.bl.bl.br),
			(0,self.fg.bl.bl.bl)]

			
	def PartitionSMTSearch(self,smt):
		smt.DoLIA()




class Sz256_lia(SearchTemplate):
	
	def __init__(self,offset):
	
		SearchTemplate.__init__(self,8,offset)
		self.name = "Sz256_lia"
		
		self.partitions = [
			(12,self.fg.br.br.tr.tr), 
			(12,self.fg.br.br.tr.tl),
			(0,self.fg.br.br.tl.tr),
			(0,self.fg.br.br.tl.tl),
			(12,self.fg.br.bl.tr.tr),
			(12,self.fg.br.bl.tr.tl),
			(0,self.fg.br.bl.tl.tr),
			(0,self.fg.br.bl.tl.tl),
#	 
			(0,self.fg.bl.br.br.tr.tr),
			(0,self.fg.bl.br.br.tr.tl),	 
			(0,self.fg.bl.br.br.tl.tr),
			(0,self.fg.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.br.bl.tr.tr),
			(0,self.fg.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.br.bl.tl.tr),
			(0,self.fg.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.br.br.tr.tr),
			(0,self.fg.bl.bl.br.br.tr.tl),	 
			(0,self.fg.bl.bl.br.br.tl.tr),
			(0,self.fg.bl.bl.br.br.tl.tl),	 	 
			(0,self.fg.bl.bl.br.bl.tr.tr),
			(0,self.fg.bl.bl.br.bl.tr.tl),	 
			(0,self.fg.bl.bl.br.bl.tl.tr),
			(0,self.fg.bl.bl.br.bl.tl.tl),	 	 	 
#	 
			(0,self.fg.bl.bl.bl.br.br.tr),
			(0,self.fg.bl.bl.bl.br.br.tl),
			(0,self.fg.bl.bl.bl.br.bl.tr),
			(0,self.fg.bl.bl.bl.br.bl.tl),	 
#	 
			(0,self.fg.bl.bl.bl.bl.br),
			(0,self.fg.bl.bl.bl.bl.bl)]
			
	def PartitionSMTSearch(self,smt):
		smt.DoLIA()



